@@ -1,12 +1,14 @@
-// Copyright (C) 2005, 2008 International Business Machines and others.
+// Copyright (C) 2005, 2010 International Business Machines and others.
 // All Rights Reserved.
 // This code is published under the Common Public License.
 //
-// $Id: IpWsmpSolverInterface.cpp 1256 2008-06-26 17:33:33Z andreasw $
+// $Id: IpWsmpSolverInterface.cpp 1791 2010-10-30 17:56:17Z andreasw $
 //
 // Authors:  Andreas Waechter                 IBM    2006-01-04
 //
 
+
+
 #include "IpoptConfig.h"
 #include "IpWsmpSolverInterface.hpp"
 
@@ -20,6 +22,16 @@
 # endif
 #endif
 
+#ifdef HAVE_CSTDIO
+# include <cstdio>
+#else
+# ifdef HAVE_STDIO_H
+#  include <stdio.h>
+# else
+#  error "don't have header file for stdio"
+# endif
+#endif
+
 /** Prototypes for WSMP's subroutines */
 extern "C"
 {
@@ -34,8 +46,15 @@ extern "C"
                              ipfint* MRP, ipfint* IPARM,
                              double* DPARM);
   void F77_FUNC_(wsmp_clear,WSMP_CLEAR)(void);
+
+#ifdef PARDISO_MATCHING_PREPROCESS
+  void smat_reordering_pardiso_wsmp_(const ipfint* N, const ipfint* ia, const ipfint* ja, const double* a_, ipfint* a2, ipfint* ja2,  double* a2_,
+                                     ipfint* perm2,  double* scale2, ipfint* tmp2_, ipfint preprocess );
+#endif
+
 }
 
+
 namespace Ipopt
 {
 #if COIN_IPOPT_VERBOSITY > 0
@@ -45,6 +64,15 @@ namespace Ipopt
   WsmpSolverInterface::WsmpSolverInterface()
       :
       a_(NULL),
+
+#ifdef PARDISO_MATCHING_PREPROCESS
+      ia2(NULL),
+      ja2(NULL),
+      a2_(NULL),
+      perm2(NULL),
+      scale2(NULL),
+#endif
+
       negevals_(-1),
       initialized_(false),
 
@@ -66,6 +94,14 @@ namespace Ipopt
     // Clear WSMP's memory
     F77_FUNC_(wsmp_clear,WSMP_CLEAR)();
 
+#ifdef PARDISO_MATCHING_PREPROCESS
+    delete[] ia2;
+    delete[] ja2;
+    delete[] a2_;
+    delete[] perm2;
+    delete[] scale2;
+#endif
+
     delete[] PERM_;
     delete[] INVP_;
     delete[] MRP_;
@@ -76,18 +112,24 @@ namespace Ipopt
 
   void WsmpSolverInterface::RegisterOptions(SmartPtr<RegisteredOptions> roptions)
   {
-    roptions->AddLowerBoundedIntegerOption(
+    roptions->AddIntegerOption(
       "wsmp_num_threads",
       "Number of threads to be used in WSMP",
-      1, 1,
+      1,
       "This determines on how many processors WSMP is running on.  This option "
       "is only available if Ipopt has been compiled with WSMP.");
     roptions->AddBoundedIntegerOption(
       "wsmp_ordering_option",
-      "Determines how ordering is done in WSMP",
+      "Determines how ordering is done in WSMP (IPARM(16)",
       -2, 3, 1,
       "This corresponds to the value of WSSMP's IPARM(16).  This option is "
       "only available if Ipopt has been compiled with WSMP.");
+    roptions->AddBoundedIntegerOption(
+      "wsmp_ordering_option2",
+      "Determines how ordering is done in WSMP (IPARM(20)",
+      0, 3, 1,
+      "This corresponds to the value of WSSMP's IPARM(20).  This option is "
+      "only available if Ipopt has been compiled with WSMP.");
     roptions->AddBoundedNumberOption(
       "wsmp_pivtol",
       "Pivot tolerance for the linear solver WSMP.",
@@ -124,13 +166,36 @@ namespace Ipopt
       "If non-negative, this option determines the iteration in which all "
       "matrices given to WSMP are written to files.  This option is only "
       "available if Ipopt has been compiled with WSMP.");
+    roptions->AddStringOption2(
+      "wsmp_skip_inertia_check",
+      "Always pretent inertia is correct.",
+      "no",
+      "no", "check inertia",
+      "yes", "skip inertia check",
+      "Setting this option to \"yes\" essentially disables inertia check. "
+      "This option makes the algorithm non-robust and easily fail, but it "
+      "might give some insight into the necessity of inertia control.");
+    roptions->AddStringOption2(
+      "wsmp_no_pivoting",
+      "Use the static pivoting option of WSMP.",
+      "no",
+      "no", "use the regular version",
+      "yes", "use static pivoting",
+      "Setting this option to \"yes\" means that WSMP instructed not to do "
+      "pivoting.  This works only in certain situations (when the Hessian "
+      "block is known to be positive definite or when we are using L-BFGS). "
+      "It can also lead to a lot of fill-in.");
   }
 
   bool WsmpSolverInterface::InitializeImpl(const OptionsList& options,
       const std::string& prefix)
   {
     options.GetIntegerValue("wsmp_num_threads", wsmp_num_threads_, prefix);
-    options.GetIntegerValue("wsmp_ordering_option", wsmp_ordering_option_,
+    Index wsmp_ordering_option;
+    options.GetIntegerValue("wsmp_ordering_option", wsmp_ordering_option,
+                            prefix);
+    Index wsmp_ordering_option2;
+    options.GetIntegerValue("wsmp_ordering_option2", wsmp_ordering_option2,
                             prefix);
     options.GetNumericValue("wsmp_pivtol", wsmp_pivtol_, prefix);
     if (options.GetNumericValue("wsmp_pivtolmax", wsmp_pivtolmax_, prefix)) {
@@ -146,12 +211,18 @@ namespace Ipopt
     options.GetIntegerValue("wsmp_scaling", wsmp_scaling_, prefix);
     options.GetIntegerValue("wsmp_write_matrix_iteration",
                             wsmp_write_matrix_iteration_, prefix);
+    options.GetBoolValue("wsmp_skip_inertia_check",
+                         skip_inertia_check_, prefix);
+    options.GetBoolValue("wsmp_no_pivoting",
+                         wsmp_no_pivoting_, prefix);
 
     // Reset all private data
     dim_=0;
     initialized_=false;
+    printed_num_threads_ = false;
     pivtol_changed_ = false;
     have_symbolic_factorization_ = false;
+    factorizations_since_recomputed_ordering_ = -1;
     delete[] a_;
     a_ = NULL;
     delete[] PERM_;
@@ -161,6 +232,23 @@ namespace Ipopt
     delete[] MRP_;
     MRP_ = NULL;
 
+#ifdef PARDISO_MATCHING_PREPROCESS
+    delete[] ia2;
+    ia2 = NULL;
+
+    delete[] ja2;
+    ja2 = NULL;
+
+    delete[] a2_;
+    a2_ = NULL;
+
+    delete[] perm2;
+    perm2 = NULL;
+
+    delete[] scale2;
+    scale2 = NULL;
+#endif
+
     // Set the number of threads
     ipfint NTHREADS = wsmp_num_threads_;
     F77_FUNC(wsetmaxthrds,WSETMAXTHRDS)(&NTHREADS);
@@ -176,12 +264,16 @@ namespace Ipopt
     F77_FUNC(wssmp,WSSMP)(&idmy, &idmy, &idmy, &ddmy, &ddmy, &idmy,
                           &idmy, &ddmy, &idmy, &idmy, &ddmy, &idmy,
                           &idmy, IPARM_, DPARM_);
-    IPARM_[14] = 0; // no restrictions on pivoting (ignored for
-    // Bunch-Kaufman)
-    IPARM_[15] = wsmp_ordering_option_; // ordering option
-    IPARM_[17] = 1; // use local minimum fill-in ordering
-    IPARM_[19] = 1; // for ordering in IP methods?
-    IPARM_[30] = 2; // want L D L^T factorization with diagonal
+    IPARM_[15] = wsmp_ordering_option; // ordering option
+    IPARM_[17] = 0; // use local minimum fill-in ordering
+    IPARM_[19] = wsmp_ordering_option2; // for ordering in IP methods?
+    if (wsmp_no_pivoting_) {
+      IPARM_[30] = 1; // want L D L^T factorization with diagonal no pivoting
+      IPARM_[26] = 1;
+    }
+    else {
+      IPARM_[30] = 2; // want L D L^T factorization with diagonal with pivoting
+    }
     // pivoting (Bunch/Kaufman)
     //IPARM_[31] = 1; // need D to see where first negative eigenvalue occurs
     //                   if we change this, we need DIAG arguments below!
@@ -213,6 +305,12 @@ namespace Ipopt
     DBG_ASSERT(!check_NegEVals || ProvidesInertia());
     DBG_ASSERT(initialized_);
 
+    if (!printed_num_threads_) {
+      Jnlst().Printf(J_ITERSUMMARY, J_LINEAR_ALGEBRA,
+                     "  -- WSMP is working with %d thread%s.\n", IPARM_[32],
+                     IPARM_[32]==1 ? "" : "s");
+      printed_num_threads_ = true;
+    }
     // check if a factorization has to be done
     if (new_matrix || pivtol_changed_) {
       pivtol_changed_ = false;
@@ -245,6 +343,7 @@ namespace Ipopt
   {
     DBG_START_METH("WsmpSolverInterface::InitializeStructure",dbg_verbosity);
     dim_ = dim;
+    nonzeros_ = nonzeros;
 
     // Make space for storing the matrix elements
     delete[] a_;
@@ -278,7 +377,8 @@ namespace Ipopt
   ESymSolverStatus
   WsmpSolverInterface::InternalSymFact(
     const Index* ia,
-    const Index* ja)
+    const Index* ja,
+    Index numberOfNegEVals)
   {
     if (HaveIpData()) {
       IpData().TimingStats().LinearSystemSymbolicFactorization().Start();
@@ -295,20 +395,64 @@ namespace Ipopt
     INVP_ = new ipfint[dim_];
     MRP_ = new ipfint[dim_];
 
-    // Call WSSMP for ordering and symbolic factorization
     ipfint N = dim_;
+
+#ifdef PARDISO_MATCHING_PREPROCESS
+
+    delete[] ia2;
+    ia2 = NULL;
+
+    delete[] ja2;
+    ja2 = NULL;
+
+    delete[] a2_;
+    a2_ = NULL;
+
+    delete[] perm2;
+    perm2 = NULL;
+
+    delete[] scale2;
+    scale2 = NULL;
+
+    ia2    = new ipfint[N+1];
+    ja2    = new ipfint[nonzeros_];
+    a2_    = new double[nonzeros_];
+    perm2  = new ipfint[N];
+    scale2 = new double[N];
+    ipfint* tmp2_  = new ipfint[N];
+
+    smat_reordering_pardiso_wsmp_(&N, ia, ja, a_, ia2, ja2, a2_, perm2,
+                                  scale2, tmp2_, 0);
+
+    delete[] tmp2_;
+
+#endif
+
+
+    // Call WSSMP for ordering and symbolic factorization
     ipfint NAUX = 0;
     IPARM_[1] = 1; // ordering
     IPARM_[2] = 2; // symbolic factorization
     ipfint idmy;
     double ddmy;
+
+    if (wsmp_no_pivoting_) {
+      IPARM_[14] = dim_ - numberOfNegEVals; // CHECK
+      Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
+                     "Restricting WSMP static pivot sequence with IPARM(15) = %d\n", IPARM_[14]);
+    }
+
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Calling WSSMP-1-2 for ordering and symbolic factorization at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
+                   "Calling WSSMP-1-2 for ordering and symbolic factorization at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
+#ifdef PARDISO_MATCHING_PREPROCESS
+    F77_FUNC(wssmp,WSSMP)(&N,  ia2,  ja2,  a2_, &ddmy, PERM_, INVP_,
+#else
     F77_FUNC(wssmp,WSSMP)(&N, ia, ja, a_, &ddmy, PERM_, INVP_,
+#endif
                           &ddmy, &idmy, &idmy, &ddmy, &NAUX, MRP_,
                           IPARM_, DPARM_);
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Done with WSSMP-1-2 for ordering and symbolic factorization at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
+                   "Done with WSSMP-1-2 for ordering and symbolic factorization at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
 
     Index ierror = IPARM_[63];
     if (ierror!=0) {
@@ -365,8 +509,8 @@ namespace Ipopt
     if (iter_count == wsmp_write_matrix_iteration_) {
       matrix_file_number_++;
       char buf[256];
-      sprintf(buf, "wsmp_matrix_%d_%d.dat", iter_count,
-              matrix_file_number_);
+      Snprintf(buf, 255, "wsmp_matrix_%d_%d.dat", iter_count,
+               matrix_file_number_);
       Jnlst().Printf(J_SUMMARY, J_LINEAR_ALGEBRA,
                      "Writing WSMP matrix into file %s.\n", buf);
       FILE* fp = fopen(buf, "w");
@@ -385,7 +529,7 @@ namespace Ipopt
     // Check if we have to do the symbolic factorization and ordering
     // phase yet
     if (!have_symbolic_factorization_) {
-      ESymSolverStatus retval = InternalSymFact(ia, ja);
+      ESymSolverStatus retval = InternalSymFact(ia, ja, numberOfNegEVals);
       if (retval != SYMSOLVER_SUCCESS) {
         return retval;
       }
@@ -406,11 +550,15 @@ namespace Ipopt
     double ddmy;
 
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Calling WSSMP-3-3 for numerical factorization at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
-    F77_FUNC(wssmp,WSSMP)(&N, ia, ja, a_, &ddmy, PERM_, INVP_, &ddmy, &idmy,
+                   "Calling WSSMP-3-3 for numerical factorization at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
+#ifdef PARDISO_MATCHING_PREPROCESS
+    F77_FUNC(wssmp,WSSMP)(&N, ia2, ja2, a2_, &ddmy, PERM_, INVP_, &ddmy, &idmy,
+#else
+    F77_FUNC(wssmp,WSSMP)(&N,  ia,  ja,  a_, &ddmy, PERM_, INVP_, &ddmy, &idmy,
+#endif
                           &idmy, &ddmy, &NAUX, MRP_, IPARM_, DPARM_);
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Done with WSSMP-3-3 for numerical factorization at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
+                   "Done with WSSMP-3-3 for numerical factorization at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
 
     const Index ierror = IPARM_[63];
     if (ierror > 0) {
@@ -442,8 +590,11 @@ namespace Ipopt
                    "Number of nonzeros in WSSMP after factorization IPARM(24) = %d\n",
                    IPARM_[23]);
 
-    negevals_ = IPARM_[21]; // Number of negative eigenvalues
-    // determined during factorization
+    if (factorizations_since_recomputed_ordering_ != -1) {
+      factorizations_since_recomputed_ordering_++;
+    }
+
+    negevals_ = IPARM_[21]; // Number of negative eigenvalues determined during factorization
 
     // Check whether the number of negative eigenvalues matches the requested
     // count
@@ -451,10 +602,18 @@ namespace Ipopt
       Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
                      "Wrong inertia: required are %d, but we got %d.\n",
                      numberOfNegEVals, negevals_);
-      if (HaveIpData()) {
-        IpData().TimingStats().LinearSystemFactorization().End();
+      if (skip_inertia_check_) {
+        Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
+                       "  But wsmp_skip_inertia_check is set.  Ignore inertia.\n");
+        IpData().Append_info_string("IC ");
+        negevals_ = numberOfNegEVals;
+      }
+      else {
+        if (HaveIpData()) {
+          IpData().TimingStats().LinearSystemFactorization().End();
+        }
+        return SYMSOLVER_WRONG_INERTIA;
       }
-      return SYMSOLVER_WRONG_INERTIA;
     }
 
     if (HaveIpData()) {
@@ -487,15 +646,31 @@ namespace Ipopt
     IPARM_[5] = 1;
     DPARM_[5] = 1e-12;
 
-    ipfint idmy;
     double ddmy;
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Calling WSSMP-4-5 for backsolve at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
+                   "Calling WSSMP-4-5 for backsolve at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
+
+#ifdef PARDISO_MATCHING_PREPROCESS
+    double* X = new double[nrhs*N];
+
+    // Initialize solution with zero and save right hand side
+    for (int i = 0; i < nrhs*N; i++) {
+      X[perm2[i]] = scale2[i] * rhs_vals[i];
+    }
+    F77_FUNC(wssmp,WSSMP)(&N, ia, ja, a_, &ddmy, PERM_, INVP_,
+                          X, &LDB, &NRHS, &ddmy, &NAUX,
+                          MRP_, IPARM_, DPARM_);
+    for (int i = 0; i < N; i++) {
+      rhs_vals[i] = scale2[i]*X[perm2[i]];
+    }
+#else
     F77_FUNC(wssmp,WSSMP)(&N, ia, ja, a_, &ddmy, PERM_, INVP_,
                           rhs_vals, &LDB, &NRHS, &ddmy, &NAUX,
                           MRP_, IPARM_, DPARM_);
+#endif
+
     Jnlst().Printf(J_MOREDETAILED, J_LINEAR_ALGEBRA,
-                   "Done with WSSMP-4-5 for backsolve at cpu time %10.3f (wall %10.3f).\n", TimedTask::CpuTime(), WallclockTime());
+                   "Done with WSSMP-4-5 for backsolve at cpu time %10.3f (wall %10.3f).\n", CpuTime(), WallclockTime());
     if (HaveIpData()) {
       IpData().TimingStats().LinearSystemBackSolve().End();
     }
@@ -516,6 +691,9 @@ namespace Ipopt
                    "Number of iterative refinement steps in WSSMP: %d\n",
                    IPARM_[5]);
 
+#ifdef PARDISO_MATCHING_PREPROCESS
+    delete [] X;
+#endif
 
     return SYMSOLVER_SUCCESS;
   }
@@ -530,13 +708,24 @@ namespace Ipopt
   bool WsmpSolverInterface::IncreaseQuality()
   {
     DBG_START_METH("WsmpSolverInterface::IncreaseQuality",dbg_verbosity);
+
+    if (factorizations_since_recomputed_ordering_ == -1 ||
+        factorizations_since_recomputed_ordering_ > 2) {
+      DPARM_[14] = 1.0;
+      pivtol_changed_ = true;
+      IpData().Append_info_string("RO ");
+      factorizations_since_recomputed_ordering_ = 0;
+      Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
+                     "Triggering WSMP's recomputation of the ordering for next factorization.\n");
+      return true;
+    }
     if (wsmp_pivtol_ == wsmp_pivtolmax_) {
       return false;
     }
     pivtol_changed_ = true;
 
     Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
-                   "Indreasing pivot tolerance for WSMP from %7.2e ",
+                   "Increasing pivot tolerance for WSMP from %7.2e ",
                    wsmp_pivtol_);
     wsmp_pivtol_ = Min(wsmp_pivtolmax_, pow(wsmp_pivtol_,0.75));
     Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
@@ -559,8 +748,11 @@ namespace Ipopt
 
     c_deps.clear();
 
+    ASSERT_EXCEPTION(!wsmp_no_pivoting_, OPTION_INVALID,
+                     "WSMP dependency detection does not work without pivoting.");
+
     if (!have_symbolic_factorization_) {
-      ESymSolverStatus retval = InternalSymFact(ia, ja);
+      ESymSolverStatus retval = InternalSymFact(ia, ja, 0);
       if (retval != SYMSOLVER_SUCCESS) {
         return retval;
       }
@@ -577,11 +769,16 @@ namespace Ipopt
     ipfint idmy;
     double ddmy;
 
+#ifdef PARDISO_MATCHING_PREPROCESS
+    F77_FUNC(wssmp,WSSMP)(&N, ia2, ja2, a2_, &ddmy, PERM_, INVP_, &ddmy, &idmy,
+                          &idmy, &ddmy, &NAUX, MRP_, IPARM_, DPARM_);
+#else
     F77_FUNC(wssmp,WSSMP)(&N, ia, ja, a_, &ddmy, PERM_, INVP_, &ddmy, &idmy,
                           &idmy, &ddmy, &NAUX, MRP_, IPARM_, DPARM_);
+#endif
+
     const Index ierror = IPARM_[63];
     if (ierror == 0) {
-      Index ndegen = IPARM_[20];
       int ii = 0;
       for (int i=0; i<N; i++) {
         if (MRP_[i] == -1) {
@@ -589,7 +786,7 @@ namespace Ipopt
           ii++;
         }
       }
-      DBG_ASSERT(ii == ndegen);
+      DBG_ASSERT(ii == IPARM_[20]);
     }
     if (ierror > 0) {
       Jnlst().Printf(J_DETAILED, J_LINEAR_ALGEBRA,
